/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001-2004 - Ecole des Mines de Paris
 *
 *  Auteur       : Jose Marcio Martins da Cruz
 *                 Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique   :
 *  Creation     : Thu Feb 10 21:19:57 CET 2005
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */


#include <j-sys.h>
#include <j-chkmail.h>
#include <j-grey-client.h>


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */

#define  DEBUG_LEVEL   12

static char        *greysock = "inet:2012@localhost";

typedef struct greychan_T greychan_T;

struct greychan_T
{
  bool                ok;
  int                 socktype;

  int                 inetport;
  char               *inethost;

  char               *sockname;

  int                 sd;

  time_t              recv_timeout;
  time_t              send_timeout;
};

static greychan_T   gChan = { FALSE, AF_UNSPEC, 0, NULL, NULL, -1, 2000, 2000 };

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static int          disconnect2greyserver(int sd);

#define DISCONNECT_FROM_SERVER(x)		\
  {						\
    disconnect2greyserver(x);			\
    x = -1;					\
  }

static void         clear_errors();

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
#define             SRV_ERR_MAX     8

#define             SRV_MIN_ERR_DELAY   (1 MINUTES)
#define             SRV_MAX_ERR_DELAY   (32 MINUTES)

static time_t       error_last = (time_t) 0;
static int          error_count = 0;
static time_t       dt = SRV_MIN_ERR_DELAY;

static pthread_mutex_t err_mutex = PTHREAD_MUTEX_INITIALIZER;

static              bool
grey_server_error_inc_and_check(inc)
     bool                inc;
{
  time_t              now = time(NULL);
  bool                res = FALSE;

  if (inc)
  {
    error_count++;
    if (error_count < SRV_ERR_MAX)
      error_last = now;

    if (error_count == SRV_ERR_MAX)
    {
      dt *= 2;
      dt = MIN(dt, SRV_MAX_ERR_DELAY);
      MESSAGE_WARNING(9, "Too many errors (%d) - "
                      "disabling remote check for %d seconds", SRV_ERR_MAX, dt);

      DISCONNECT_FROM_SERVER(gChan.sd);
    }
  }

  res = (error_count < SRV_ERR_MAX) || (error_last + dt < now);

  return res;
}

static void
clear_errors()
{
  error_count = 0;
  dt = SRV_MIN_ERR_DELAY;
}

void
grey_channel_error_clear()
{
  MUTEX_LOCK(&err_mutex);

  if (error_count >= SRV_ERR_MAX && gChan.sd >= 0)
    DISCONNECT_FROM_SERVER(gChan.sd);

  clear_errors();

  MUTEX_UNLOCK(&err_mutex);
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              bool
check_channel_socket(sd)
     int                 sd;
{
  struct pollfd       pfd;
  int                 res = 0;

  if (sd < 0)
    return FALSE;

  memset(&pfd, 0, sizeof (pfd));

  pfd.fd = sd;
  pfd.events = POLLIN | POLLOUT;

  if ((res = poll(&pfd, 1, 0)) < 0)
    LOG_SYS_ERROR("poll error");

  return (res >= 0);
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              bool
grey_decode_socket(arg)
     char               *arg;
{
  char               *p = arg;
  char               *s = "unix:";

  FREE(gChan.sockname);
  FREE(gChan.inethost);
  gChan.socktype = -1;
  gChan.inetport = -1;

  if (strncasecmp(p, s, strlen(s)) == 0)
  {
    p += strlen(s);
    if (strlen(p) > 0)
    {
      if ((gChan.sockname = strdup(p)) == NULL)
      {
        LOG_SYS_ERROR("strdup(sockname) error");
        return FALSE;
      }
    }
    gChan.socktype = AF_UNIX;
    return TRUE;
  }

  s = "local:";
  if (strncasecmp(p, s, strlen(s)) == 0)
  {
    p += strlen(s);
    if (strlen(p) > 0)
    {
      if ((gChan.sockname = strdup(p)) == NULL)
      {
        LOG_SYS_ERROR("strdup(sockname) error");
        return FALSE;
      }
    }
    gChan.socktype = AF_UNIX;
    return TRUE;
  }

  s = "inet:";
  if (strncasecmp(p, s, strlen(s)) == 0)
  {
    char                tmp[16];
    int                 n;

    p += strlen(s);
    n = strspn(p, "0123456789");
    if ((n > 0) && (n < sizeof (tmp)))
    {
      strncpy(tmp, p, n);
      tmp[n] = '\0';

      gChan.inetport = atoi(tmp);

      p += n;
      if (*p == '@')
        p++;
      if (strlen(p) > 0)
        gChan.inethost = strdup(p);
      else
        gChan.inethost = strdup("localhost");
      if (gChan.inethost == NULL)
        LOG_SYS_ERROR("strdup(inethost) error");
    }

    if ((gChan.inethost == NULL) || (gChan.inetport < 0))
    {
      FREE(gChan.inethost);
      gChan.inetport = -1;
      return FALSE;
    }
    gChan.socktype = AF_INET;
    return TRUE;
  }

  return FALSE;
}



/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static int
disconnect2greyserver(sd)
     int                 sd;
{
  if (sd >= 0)
  {
    (void) sd_printf(sd, "QUIT\r\n");

#if 0
    sleep(1);
#endif
    shutdown(sd, SHUT_RDWR);
    close(sd);
  }
  return -1;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static int
connect2greyserver()
{
  int                 sd = -1;
  char               *env = NULL;

  gChan.recv_timeout = 2000;
  if ((env = getenv("GREYD_RECV_TIMEOUT")) != NULL)
  {
    long                to = str2long(env, NULL, gChan.recv_timeout);

    if (to > 100)
      gChan.recv_timeout = to;
  }
  gChan.send_timeout = 2000;
  if ((env = getenv("GREYD_SEND_TIMEOUT")) != NULL)
  {
    long                to = str2long(env, NULL, gChan.send_timeout);

    if (to > 100)
      gChan.send_timeout = to;
  }

  if (gChan.socktype == AF_INET)
  {
    struct sockaddr_in  his_sock;
    struct hostent     *hp;
    int                 to;

    to = cf_get_int(CF_GREY_CONNECT_TIMEOUT);
    if (to < 1)
      to = 1;

    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      LOG_SYS_ERROR("AF_INET : socket");
      return -1;
    }

    /* adresse destinataire XXX */
    if ((hp = gethostbyname(gChan.inethost)) == NULL)
    {
      LOG_SYS_ERROR("gethostbyname(%s)", STRNULL(gChan.inethost, "NULL"));
      return -1;
    }

    memcpy((char *) &his_sock.sin_addr, hp->h_addr, hp->h_length);
    his_sock.sin_family = AF_INET;
    his_sock.sin_port = htons(gChan.inetport);

    if (log_level >= DEBUG_LEVEL)
      log_sock_addr(&his_sock);

    /* emission sur sd vers his_sock d'un message de taille size */
    if (connect_timed(sd, (struct sockaddr *) &his_sock, sizeof (his_sock), to)
        != 0)
    {
      LOG_SYS_ERROR("connect error (%s:%d)", STRNULL(gChan.inethost, "NULL"),
                    gChan.inetport);
      DISCONNECT_FROM_SERVER(sd);
    }
    return sd;
  }

  if (gChan.socktype == AF_UNIX)
  {
    struct sockaddr_un  his_sock;

    if ((gChan.sockname != NULL) && (strlen(gChan.sockname) == 0))
    {
      LOG_SYS_ERROR("AF_UNIX : No sockname given...");
      return -1;
    }

    if ((sd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
    {
      LOG_SYS_ERROR("AF_UNIX : socket");
      return -1;
    }

    /* unlink(sockname); */
    memset(&his_sock, 0, sizeof (his_sock));
    his_sock.sun_family = AF_UNIX;

    strncpy(his_sock.sun_path, gChan.sockname, strlen(gChan.sockname) + 1);

    /* emission sur sd vers his_sock d'un message de taille size */
    if (connect(sd, (struct sockaddr *) &his_sock, sizeof (his_sock)) != 0)
    {
      LOG_SYS_ERROR("connect error (%s)", STRNULL(gChan.sockname, "NULL"));
      DISCONNECT_FROM_SERVER(sd);
    }

    return sd;
  }

  MESSAGE_ERROR(8, "Family socket unknown... %d", gChan.socktype);

  return -1;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              bool
grey_channel_check()
{
  char               *scan_arg = NULL;
  static bool         ok = FALSE;

  if (!ok)
  {
    atexit(remote_grey_quit);
    ok = TRUE;
  }

  /*
   ** Check errors
   ** if :
   **   ok - continue
   **   ko - return false
   */
  if (!grey_server_error_inc_and_check(FALSE))
    return FALSE;

  if (!gChan.ok)
  {

    if ((scan_arg = cf_get_str(CF_GREY_SOCKET)) != NULL)
      greysock = scan_arg;

    gChan.ok = grey_decode_socket(greysock);

    MESSAGE_INFO(DEBUG_LEVEL, "SOCK            : %s",
                 STRNULL(greysock, "NULL"));
    MESSAGE_INFO(DEBUG_LEVEL, "SOCKTYPE        : %d", gChan.socktype);
    MESSAGE_INFO(DEBUG_LEVEL, "SOCKNAME        : %s",
                 STRNULL(gChan.sockname, "NULL"));
    MESSAGE_INFO(DEBUG_LEVEL, "INETHOST        : %s",
                 STRNULL(gChan.inethost, "NULL"));
    MESSAGE_INFO(DEBUG_LEVEL, "INETPORT        : %d", gChan.inetport);
    MESSAGE_INFO(DEBUG_LEVEL, "INIT OK         : %d %s", gChan.ok,
                 STRBOOL(gChan.ok, "TRUE", "FALSE"));
  }

  if (!gChan.ok)
  {
    (void) grey_server_error_inc_and_check(TRUE);

    return FALSE;
  }

  if (gChan.sd >= 0 && !check_channel_socket(gChan.sd))
    DISCONNECT_FROM_SERVER(gChan.sd);

  if (gChan.sd < 0)
  {
    time_t              to = 10;

    gChan.sd = connect2greyserver();

    if (gChan.sd >= 0)
    {
      while (jfd_ready(gChan.sd, J_SOCK_READ, to) == J_SOCK_READY)
      {
        size_t              sz;
        char                buf[256];

        to = 10;
        sz = recvfrom(gChan.sd, buf, sizeof (buf), 0, NULL, NULL);
        buf[sz] = '\0';
        if (sz > 0)
          MESSAGE_INFO(DEBUG_LEVEL, "Connected to j-greyd : %5d %s", sz, buf);
        else
          break;
      }
    }
  }

  /* XXX if gChan.sd < 0 -> increment error count */
  if (gChan.sd >= 0)
    clear_errors();
  else
    (void) grey_server_error_inc_and_check(TRUE);

  return gChan.sd >= 0;
}


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
int
remote_grey_check(ip, from, to, hostname)
     char               *ip;
     char               *from;
     char               *to;
     char               *hostname;
{
  int                 result = GREY_ERROR;
  char                buf[4096];
  char                line[1024];
  char               *p;
  size_t              sz;
  time_t              tout = 2000;

  MUTEX_LOCK(&mutex);

  if (!grey_channel_check())
    goto fin;

  MESSAGE_INFO(DEBUG_LEVEL, "Checking %s %s %s %s", ip, from, to, hostname);

#if _FFR_FLUSH_GREYD
  /* empty input buffer before asking something */
  if (!flush_read_socket(gChan.sd))
  {
    /* log errors and ... */
  }
#endif

  if (!sd_printf(gChan.sd, "GREYCHECK %s %s %s %s\r\n", ip, STRNULL(from, "-"),
                 STRNULL(to, "-"), STRNULL(hostname, "-")))
  {
    DISCONNECT_FROM_SERVER(gChan.sd);
    goto fin;
  }

  tout = gChan.recv_timeout;
  memset(buf, 0, sizeof (buf));
  p = buf;
  sz = sizeof (buf);
  while (jfd_ready(gChan.sd, J_SOCK_READ, tout) == J_SOCK_READY)
  {
    size_t              n;

    n = recvfrom(gChan.sd, p, sz - 1, 0, NULL, NULL);

    if (n > 0)
    {
      tout = 10;
      p += n;
      sz -= n;
      *p = '\0';

      if (n < 1)
        break;
    }

    if (n < 0)
    {
      if (errno == EINTR)
        continue;

      /* an error occured */
      LOG_SYS_ERROR("recvfrom error");
      DISCONNECT_FROM_SERVER(gChan.sd);

      goto fin;
    }

    if (n == 0)
    {
      /* an error occured */
      LOG_MSG_ERROR("greyd server performed an orderly shutdown");

#if 1
      grey_server_error_inc_and_check(TRUE);
#else
      DISCONNECT_FROM_SERVER(gChan.sd);
#endif

      goto fin;
    }
  }

  clear_errors();

  if (strlen(buf) > 0)
  {
    p = buf;

    while ((sz = buf_get_line(line, sizeof (line), p, strlen(p))) > 0)
    {
      int                 argc, code = GREY_ERROR;
      char               *argv[16];

      MESSAGE_INFO(DEBUG_LEVEL, "LINE IN : %s", line);

      p += sz;
      if (!strexpr(line, "GREYCHECK ANSWER", NULL, NULL, TRUE))
        continue;

      argc = str2tokens(line, 16, argv, " ");
      if (argc > 0)
        code = str2ulong(argv[0], NULL, GREY_OK);
      switch (code)
      {
        case GREY_OK:
        case GREY_WAIT:
        case GREY_ERROR:
        case GREY_REJECT:
          result = code;
          break;
        default:
          result = GREY_ERROR;
      }
    }
  }

fin:
  MUTEX_UNLOCK(&mutex);

  if (result == GREY_ERROR)
    remote_grey_quit();

  return result;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
int
remote_grey_validate(ip, from, to, hostname)
     char               *ip;
     char               *from;
     char               *to;
     char               *hostname;
{
  int                 result = GREY_WAIT;
  char                buf[4096];
  char                line[1024];
  char               *p;
  size_t              sz;
  time_t              tout = 100;

  MUTEX_LOCK(&mutex);

  if (!grey_channel_check())
    goto fin;

#if _FFR_FLUSH_GREYD
  /* empty input buffer before asking something */
  if (!flush_read_socket(gChan.sd))
  {
    /* log errors and ... */
  }
#endif

  if (!sd_printf(gChan.sd, "GREYVALID %s %s %s %s\r\n", ip, STRNULL(from, "-"),
                 STRNULL(to, "-"), STRNULL(hostname, "-")))
  {
    DISCONNECT_FROM_SERVER(gChan.sd);
    goto fin;
  }

  memset(buf, 0, sizeof (buf));
  p = buf;
  sz = sizeof (buf);
  while (jfd_ready(gChan.sd, J_SOCK_READ, tout) == J_SOCK_READY)
  {
    size_t              n;

    n = recvfrom(gChan.sd, p, sz - 1, 0, NULL, NULL);

    if (n > 0)
    {
      tout = 10;
      p += n;
      sz -= n;
      *p = '\0';

      if (sz < 1)
        break;
    }

    if (n < 0)
    {
      if (errno == EINTR)
        continue;

      /* an error occured */
      LOG_SYS_ERROR("recvfrom error");
      DISCONNECT_FROM_SERVER(gChan.sd);

      goto fin;
    }

    if (n == 0)
    {
      /* an error occured */
      LOG_MSG_ERROR("greyd server performed an orderly shutdown");
      DISCONNECT_FROM_SERVER(gChan.sd);

      goto fin;
    }
  }

  if (strlen(buf) > 0)
  {
    p = buf;

    while ((sz = buf_get_line(line, sizeof (line), p, strlen(p))) > 0)
    {
      int                 argc, code = GREY_ERROR;
      char               *argv[16];

      MESSAGE_INFO(DEBUG_LEVEL, "LINE IN : %s", line);

      p += sz;
      if (!strexpr(line, "GREYVALID ANSWER", NULL, NULL, TRUE))
        continue;

      argc = str2tokens(line, 16, argv, " ");
      if (argc > 0)
        code = str2ulong(argv[0], NULL, GREY_OK);
      switch (code)
      {
        case GREY_OK:
        case GREY_WAIT:
        case GREY_ERROR:
        case GREY_REJECT:
          result = code;
          break;
        default:
          result = GREY_ERROR;
      }
    }
  }

fin:
  MUTEX_UNLOCK(&mutex);

  if (result == GREY_ERROR)
    remote_grey_quit();

  return result;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
void
remote_grey_quit()
{
  MUTEX_LOCK(&mutex);

  DISCONNECT_FROM_SERVER(gChan.sd);

  clear_errors();

  MUTEX_UNLOCK(&mutex);
}


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              bool
flush_read_socket(sd)
     int                 sd;
{
  bool                result = TRUE;
  char                buf[256];

  /* empty input buffer before asking something */
  for (;;)
  {
    int                 r;
    size_t              sz;

    r = jfd_ready(sd, J_SOCK_READ, 5);

    if (r == J_SOCK_ERROR)
    {
      LOG_SYS_ERROR("error");
      DISCONNECT_FROM_SERVER(sd);
      result = FALSE;
      goto fin;
    }

    if (r != J_SOCK_READY)
      break;

    sz = recvfrom(sd, buf, sizeof (buf), 0, NULL, NULL);
    if (sz > 0)
      continue;

    if (sz < 0)
      LOG_SYS_ERROR("recvfrom error");
    if (sz == 0)
      LOG_MSG_ERROR("greyd server performed an orderly shutdown");

    /* an error occured */
    DISCONNECT_FROM_SERVER(sd);
    result = FALSE;
    break;
  }
fin:
  return result;
}
