/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */


#include <j-sys.h>

#include "j-chkmail.h"

#include "j-filter.h"
#include "j-spool.h"

#define SWAP_PTR(a,b)         do { void *c = a; a = b; b = c;} while (0)

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */

bool                spool_file_create(CTXPRIV_T *);
bool                spool_file_write(CTXPRIV_T *, char *, size_t);
bool                spool_file_close(CTXPRIV_T *);
bool                spool_file_forget(CTXPRIV_T *);


/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
bool
spool_file_create(priv)
     CTXPRIV_T          *priv;
{
  char               *p;
  char                fname[256];

  int                 fd;

  if (priv->fd >= 0)
    return TRUE;

  priv->fp_open = FALSE;

  if (priv->fname != NULL)
    LOG_MSG_WARNING("fd closed, but fname not NULL");
  FREE(priv->fname);

  if ((p = cf_get_str(CF_SPOOLDIR)) == NULL)
    p = J_SPOOLDIR;

  if (cf_get_int(CF_CLUSTER) == OPT_YES)
  {
    char               *server = NULL;

    server = STRNULL(priv->mailserver, "SERVER");

    snprintf(fname, sizeof (fname), "%s/%s.%s.%04X", p, server,
             CONNID_STR(priv->id), priv->nb_msgs);

  } else
  {
    snprintf(fname, sizeof (fname), "%s/%s.%04X", p, CONNID_STR(priv->id),
             priv->nb_msgs);
  }

  fd = open(fname, O_WRONLY | O_APPEND | O_CREAT, 0640);

  if (fd < 0)
  {
    LOG_SYS_ERROR("can't create spool file (%s)", fname);
    fd = -1;
    return FALSE;
  }

  if (fchmod(fd, S_IRUSR | S_IWUSR) != 0)
    LOG_SYS_ERROR("can't change spool file rights (%s)", fname);

  priv->fd = fd;
  priv->fp_open = TRUE;

  /* XXX lock spool file */
  /*(void ) file_lock(fd); */

  if ((priv->fname = strdup(fname)) == NULL)
    LOG_SYS_ERROR("strdup(%s) error", fname);

  /* Let's add a fake From line */
  if (cf_get_int(CF_QUARANTINE_ADD_FROM_LINE) == OPT_YES)
  {
    char                s[256];
    char                t[256];
    time_t              tid = CONNID_INT(priv->id);

#if HAVE_CTIME_R
#ifdef _POSIX_PTHREAD_SEMANTICS
    ctime_r(&tid, t);
#else
    ctime_r(&tid, t, sizeof (t));
#endif
#elif HAVE_CTIME
    strlcpy(t, ctime(&tid, sizeof (t));
#else
    strlcpy(t, "-", sizeof (t));
#endif
    if ((p = strchr(t, '\r')) != NULL)
      *p = '\0';
    if ((p = strchr(t, '\n')) != NULL)
      *p = '\0';

    if (priv->env_from != NULL)
      snprintf(s, sizeof (s) - 3, "From %s %s", priv->env_from, t);
    else
      snprintf(s, sizeof (s) - 3, "From %s %s", "<unknown>", t);
#if defined(_FFR_CLEAN_MSG_BUF)
    strcat(s, "\n");
#else
    strcat(s, CRLF);
#endif
    if (write(priv->fd, s, strlen(s)) != strlen(s))
      LOG_SYS_ERROR("can't write fake From: line to %s", priv->fname);
  }

  return TRUE;
}

/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
bool
spool_file_write(priv, buf, size)
     CTXPRIV_T          *priv;
     char               *buf;
     size_t              size;
{
  if (priv->fd < 0)
    return TRUE;

  if (write(priv->fd, buf, size) != size)
  {
    LOG_SYS_ERROR("Error writing on spool file %s", priv->fname);
    return FALSE;
  }

  return TRUE;
}

/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
bool
spool_file_close(priv)
     CTXPRIV_T          *priv;
{
  if (priv->fd < 0)
    return TRUE;

  priv->fp_open = FALSE;
  if (close(priv->fd) < 0)
  {
    LOG_SYS_ERROR("error closing spool file");
    priv->fd = -1;
    return FALSE;
  }
  priv->fd = -1;

  return TRUE;
}


/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
bool
spool_file_forget(priv)
     CTXPRIV_T          *priv;
{
  char               *filename = NULL;
  char               *suffix = NULL;
  size_t              sz = 0;

  (void) spool_file_close(priv);

  if (priv->fname == NULL)
    goto fin;

  if (!priv->save_msg)
  {
    unlink(priv->fname);
    goto fin;
  }

  suffix = STREMPTY(priv->fsuffix, SUFFIX_UNKNOWN);
  sz = strlen(priv->fname) + strlen(suffix) + 4;

  filename = (char *) malloc(sz);
  if (filename != NULL)
  {
    snprintf(filename, sz, "%s%s", priv->fname, suffix);
    if (rename(priv->fname, filename) == 0)
      SWAP_PTR(filename, priv->fname);
    else
      LOG_SYS_ERROR("Error renaming quarantine file : %s", priv->fname);
  } else
    LOG_SYS_ERROR("Quarantine file name malloc error : %s", priv->fname);
  MESSAGE_INFO(11, "%-12s : quarantine file %s", CONNID_STR(priv->id),
               priv->fname);

fin:
  priv->fsuffix = NULL;
  priv->save_msg = FALSE;
  FREE(priv->fname);
  FREE(filename);

  return TRUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
bool
spool_file_is_open(priv)
     CTXPRIV_T          *priv;
{
  if (priv == NULL)
    return FALSE;

  return priv->fp_open;
}
